#include "atlas_misc.h"
#ifndef PFYDIST
   #define PFYDIST 8
#endif
#ifndef PFXDIST
   #define PFXDIST 256
#endif
#ifndef PFADIST
   #define PFADIST 256
#endif

#if !defined(ATL_3DNow) && !defined(ATL_SSE1) && \
    (PFADIST != 0 || PFXDIST != 0 || PFYDIST != 0)
   #include "atlas_prefetch.h"
#endif

#if defined(ATL_3DNow) || defined(ATL_SSE1)
   #ifndef PFIY
      #define PFIY prefetchnta
   #endif
   #ifndef PFIX
      #define PFIX prefetchnta
   #endif
   #ifndef PFIA
      #ifdef ATL_3DNow
         #define PFIA prefetchw
      #else
         #define PFIA prefetcht0
      #endif
   #endif
#endif
/*
 * X & A are prefetched in M loop PF[A,X]DIST (in bytes) ahead
 */
#if PFADIST == 0                /* flag for no prefetch */
   #define prefA(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefA(mem) __asm__ __volatile__ \
         (Mstr(PFIA) " %0" : : "m" (*(((char *)(mem))+PFADIST)))
   #else
      #if PFLVL == 2
         #define prefA(mem) ATL_pfl2W(((char*)mem)+PFADIST)
      #else
         #define prefA(mem) ATL_pfl1W(((char*)mem)+PFADIST)
      #endif
   #endif
#endif
#if PFXDIST == 0                /* flag for no prefetch */
   #define prefX(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefX(mem) __asm__ __volatile__ \
         (Mstr(PFIX) " %0" : : "m" (*(((char *)(mem))+PFXDIST)))
   #else
      #if PFLVL == 2
         #define prefX(mem) ATL_pfl2R(((char*)mem)+PFXDIST)
      #else
         #define prefX(mem) ATL_pfl1R(((char*)mem)+PFXDIST)
      #endif
   #endif
#endif
/*
 * Y is prefetched in N-loop, and always fetches next NU piece
 */
#if PFYDIST == 0                /* flag for no prefetch */
   #define prefY(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefY(mem) \
         __asm__ __volatile__ (Mstr(PFIY) " %0" : : "m" (*(((char *)(mem)))))
   #else
      #if PFLVL == 2
         #define prefY(mem) ATL_pfl2R(mem)
      #else
         #define prefY(mem) ATL_pfl1R(mem)
      #endif
   #endif
#endif
#ifndef ATL_CINT
   #define ATL_CINT const int
#endif
#ifndef ATL_INT
   #define ATL_INT int
#endif
void ATL_UGER2K(ATL_CINT M0, ATL_CINT N, const float *X, const float *Y, 
                const float *W, const float *Z, float *A, ATL_CINT lda0)
/*
 * 8x8 unrolled r2_c.
 * Extracted from ATLAS/tune/blas/gemv/atlas-l2g.base
 */
{
   ATL_CINT N8=(N/8)*8, M=M0+M0, M8=((M0/8)*8)<<1,
            lda=lda0+lda0, lda8=lda*8;
   ATL_INT j;

   for (j=N8; j; j -= 8, A += lda8, Y += 16)
   {
      float *A0=A, *A1=A0+lda, *A2=A1+lda, *A3=A2+lda, *A4=A3+lda, *A5=A4+lda,
            *A6=A5+lda, *A7=A6+lda;
      const register float ry0=Y[0], rz0=Z[0], iy0=Y[1], iz0=Z[1], ry1=Y[2],
                           rz1=Z[2], iy1=Y[3], iz1=Z[3], ry2=Y[4], rz2=Z[4],
                           iy2=Y[5], iz2=Z[5], ry3=Y[6], rz3=Z[6], iy3=Y[7],
                           iz3=Z[7], ry4=Y[8], rz4=Z[8], iy4=Y[9], iz4=Z[9],
                           ry5=Y[10], rz5=Z[10], iy5=Y[11], iz5=Z[11],
                           ry6=Y[12], rz6=Z[12], iy6=Y[13], iz6=Z[13],
                           ry7=Y[14], rz7=Z[14], iy7=Y[15], iz7=Z[15];
      ATL_INT i;
      prefY(Y+16+16-1);
      for (i=0; i < M8; i += 16)
      {
         register float ra00, ia00, ra10, ia10;
         const float rx0=X[i+0], rw0=W[i+0], ix0=X[i+1], iw0=W[i+1], rx1=X[i+2],
                     rw1=W[i+2], ix1=X[i+3], iw1=W[i+3], rx2=X[i+4], rw2=W[i+4],
                     ix2=X[i+5], iw2=W[i+5], rx3=X[i+6], rw3=W[i+6], ix3=X[i+7],
                     iw3=W[i+7], rx4=X[i+8], rw4=W[i+8], ix4=X[i+9], iw4=W[i+9],
                     rx5=X[i+10], rw5=W[i+10], ix5=X[i+11], iw5=W[i+11],
                     rx6=X[i+12], rw6=W[i+12], ix6=X[i+13], iw6=W[i+13],
                     rx7=X[i+14], rw7=W[i+14], ix7=X[i+15], iw7=W[i+15];

         prefX(X);
         ra00 = A0[i+0];
         ia00 = A0[i+0+1];
         ra10 = A0[i+0+2];
         ia10 = A0[i+0+3];
         ra00 += rx0 * ry0;
         ia00 += rx0 * iy0;
         ra10 += rx1 * ry0;
         ia10 += rx1 * iy0;
         prefA(A0);
         ra00 -= ix0 * iy0;
         ia00 += ix0 * ry0;
         ra10 -= ix1 * iy0;
         ia10 += ix1 * ry0;
         prefX(W);
         ra00 += rw0 * rz0;
         ia00 += rw0 * iz0;
         ra10 += rw1 * rz0;
         ia10 += rw1 * iz0;
         ra00 -= iw0 * iz0;
         A0[i+0] = ra00;
         ia00 += iw0 * rz0;
         A0[i+0+1] = ia00;
         ra10 -= iw1 * iz0;
         A0[i+0+2] = ra10;
         ia10 += iw1 * rz0;
         A0[i+0+3] = ia10;
         ra00 = A0[i+4];
         ia00 = A0[i+4+1];
         ra10 = A0[i+4+2];
         ia10 = A0[i+4+3];
         ra00 += rx2 * ry0;
         ia00 += rx2 * iy0;
         ra10 += rx3 * ry0;
         ia10 += rx3 * iy0;
         ra00 -= ix2 * iy0;
         ia00 += ix2 * ry0;
         ra10 -= ix3 * iy0;
         ia10 += ix3 * ry0;
         ra00 += rw2 * rz0;
         ia00 += rw2 * iz0;
         ra10 += rw3 * rz0;
         ia10 += rw3 * iz0;
         ra00 -= iw2 * iz0;
         A0[i+4] = ra00;
         ia00 += iw2 * rz0;
         A0[i+4+1] = ia00;
         ra10 -= iw3 * iz0;
         A0[i+4+2] = ra10;
         ia10 += iw3 * rz0;
         A0[i+4+3] = ia10;
         ra00 = A0[i+8];
         ia00 = A0[i+8+1];
         ra10 = A0[i+8+2];
         ia10 = A0[i+8+3];
         ra00 += rx4 * ry0;
         ia00 += rx4 * iy0;
         ra10 += rx5 * ry0;
         ia10 += rx5 * iy0;
         ra00 -= ix4 * iy0;
         ia00 += ix4 * ry0;
         ra10 -= ix5 * iy0;
         ia10 += ix5 * ry0;
         ra00 += rw4 * rz0;
         ia00 += rw4 * iz0;
         ra10 += rw5 * rz0;
         ia10 += rw5 * iz0;
         ra00 -= iw4 * iz0;
         A0[i+8] = ra00;
         ia00 += iw4 * rz0;
         A0[i+8+1] = ia00;
         ra10 -= iw5 * iz0;
         A0[i+8+2] = ra10;
         ia10 += iw5 * rz0;
         A0[i+8+3] = ia10;
         ra00 = A0[i+12];
         ia00 = A0[i+12+1];
         ra10 = A0[i+12+2];
         ia10 = A0[i+12+3];
         ra00 += rx6 * ry0;
         ia00 += rx6 * iy0;
         ra10 += rx7 * ry0;
         ia10 += rx7 * iy0;
         ra00 -= ix6 * iy0;
         ia00 += ix6 * ry0;
         ra10 -= ix7 * iy0;
         ia10 += ix7 * ry0;
         ra00 += rw6 * rz0;
         ia00 += rw6 * iz0;
         ra10 += rw7 * rz0;
         ia10 += rw7 * iz0;
         ra00 -= iw6 * iz0;
         A0[i+12] = ra00;
         ia00 += iw6 * rz0;
         A0[i+12+1] = ia00;
         ra10 -= iw7 * iz0;
         A0[i+12+2] = ra10;
         ia10 += iw7 * rz0;
         A0[i+12+3] = ia10;

         ra00 = A1[i+0];
         ia00 = A1[i+0+1];
         ra10 = A1[i+0+2];
         ia10 = A1[i+0+3];
         ra00 += rx0 * ry1;
         ia00 += rx0 * iy1;
         ra10 += rx1 * ry1;
         ia10 += rx1 * iy1;
         prefA(A1);
         ra00 -= ix0 * iy1;
         ia00 += ix0 * ry1;
         ra10 -= ix1 * iy1;
         ia10 += ix1 * ry1;
         ra00 += rw0 * rz1;
         ia00 += rw0 * iz1;
         ra10 += rw1 * rz1;
         ia10 += rw1 * iz1;
         ra00 -= iw0 * iz1;
         A1[i+0] = ra00;
         ia00 += iw0 * rz1;
         A1[i+0+1] = ia00;
         ra10 -= iw1 * iz1;
         A1[i+0+2] = ra10;
         ia10 += iw1 * rz1;
         A1[i+0+3] = ia10;
         ra00 = A1[i+4];
         ia00 = A1[i+4+1];
         ra10 = A1[i+4+2];
         ia10 = A1[i+4+3];
         ra00 += rx2 * ry1;
         ia00 += rx2 * iy1;
         ra10 += rx3 * ry1;
         ia10 += rx3 * iy1;
         ra00 -= ix2 * iy1;
         ia00 += ix2 * ry1;
         ra10 -= ix3 * iy1;
         ia10 += ix3 * ry1;
         ra00 += rw2 * rz1;
         ia00 += rw2 * iz1;
         ra10 += rw3 * rz1;
         ia10 += rw3 * iz1;
         ra00 -= iw2 * iz1;
         A1[i+4] = ra00;
         ia00 += iw2 * rz1;
         A1[i+4+1] = ia00;
         ra10 -= iw3 * iz1;
         A1[i+4+2] = ra10;
         ia10 += iw3 * rz1;
         A1[i+4+3] = ia10;
         ra00 = A1[i+8];
         ia00 = A1[i+8+1];
         ra10 = A1[i+8+2];
         ia10 = A1[i+8+3];
         ra00 += rx4 * ry1;
         ia00 += rx4 * iy1;
         ra10 += rx5 * ry1;
         ia10 += rx5 * iy1;
         ra00 -= ix4 * iy1;
         ia00 += ix4 * ry1;
         ra10 -= ix5 * iy1;
         ia10 += ix5 * ry1;
         ra00 += rw4 * rz1;
         ia00 += rw4 * iz1;
         ra10 += rw5 * rz1;
         ia10 += rw5 * iz1;
         ra00 -= iw4 * iz1;
         A1[i+8] = ra00;
         ia00 += iw4 * rz1;
         A1[i+8+1] = ia00;
         ra10 -= iw5 * iz1;
         A1[i+8+2] = ra10;
         ia10 += iw5 * rz1;
         A1[i+8+3] = ia10;
         ra00 = A1[i+12];
         ia00 = A1[i+12+1];
         ra10 = A1[i+12+2];
         ia10 = A1[i+12+3];
         ra00 += rx6 * ry1;
         ia00 += rx6 * iy1;
         ra10 += rx7 * ry1;
         ia10 += rx7 * iy1;
         ra00 -= ix6 * iy1;
         ia00 += ix6 * ry1;
         ra10 -= ix7 * iy1;
         ia10 += ix7 * ry1;
         ra00 += rw6 * rz1;
         ia00 += rw6 * iz1;
         ra10 += rw7 * rz1;
         ia10 += rw7 * iz1;
         ra00 -= iw6 * iz1;
         A1[i+12] = ra00;
         ia00 += iw6 * rz1;
         A1[i+12+1] = ia00;
         ra10 -= iw7 * iz1;
         A1[i+12+2] = ra10;
         ia10 += iw7 * rz1;
         A1[i+12+3] = ia10;

         ra00 = A2[i+0];
         ia00 = A2[i+0+1];
         ra10 = A2[i+0+2];
         ia10 = A2[i+0+3];
         ra00 += rx0 * ry2;
         ia00 += rx0 * iy2;
         ra10 += rx1 * ry2;
         ia10 += rx1 * iy2;
         prefA(A2);
         ra00 -= ix0 * iy2;
         ia00 += ix0 * ry2;
         ra10 -= ix1 * iy2;
         ia10 += ix1 * ry2;
         ra00 += rw0 * rz2;
         ia00 += rw0 * iz2;
         ra10 += rw1 * rz2;
         ia10 += rw1 * iz2;
         ra00 -= iw0 * iz2;
         A2[i+0] = ra00;
         ia00 += iw0 * rz2;
         A2[i+0+1] = ia00;
         ra10 -= iw1 * iz2;
         A2[i+0+2] = ra10;
         ia10 += iw1 * rz2;
         A2[i+0+3] = ia10;
         ra00 = A2[i+4];
         ia00 = A2[i+4+1];
         ra10 = A2[i+4+2];
         ia10 = A2[i+4+3];
         ra00 += rx2 * ry2;
         ia00 += rx2 * iy2;
         ra10 += rx3 * ry2;
         ia10 += rx3 * iy2;
         ra00 -= ix2 * iy2;
         ia00 += ix2 * ry2;
         ra10 -= ix3 * iy2;
         ia10 += ix3 * ry2;
         ra00 += rw2 * rz2;
         ia00 += rw2 * iz2;
         ra10 += rw3 * rz2;
         ia10 += rw3 * iz2;
         ra00 -= iw2 * iz2;
         A2[i+4] = ra00;
         ia00 += iw2 * rz2;
         A2[i+4+1] = ia00;
         ra10 -= iw3 * iz2;
         A2[i+4+2] = ra10;
         ia10 += iw3 * rz2;
         A2[i+4+3] = ia10;
         ra00 = A2[i+8];
         ia00 = A2[i+8+1];
         ra10 = A2[i+8+2];
         ia10 = A2[i+8+3];
         ra00 += rx4 * ry2;
         ia00 += rx4 * iy2;
         ra10 += rx5 * ry2;
         ia10 += rx5 * iy2;
         ra00 -= ix4 * iy2;
         ia00 += ix4 * ry2;
         ra10 -= ix5 * iy2;
         ia10 += ix5 * ry2;
         ra00 += rw4 * rz2;
         ia00 += rw4 * iz2;
         ra10 += rw5 * rz2;
         ia10 += rw5 * iz2;
         ra00 -= iw4 * iz2;
         A2[i+8] = ra00;
         ia00 += iw4 * rz2;
         A2[i+8+1] = ia00;
         ra10 -= iw5 * iz2;
         A2[i+8+2] = ra10;
         ia10 += iw5 * rz2;
         A2[i+8+3] = ia10;
         ra00 = A2[i+12];
         ia00 = A2[i+12+1];
         ra10 = A2[i+12+2];
         ia10 = A2[i+12+3];
         ra00 += rx6 * ry2;
         ia00 += rx6 * iy2;
         ra10 += rx7 * ry2;
         ia10 += rx7 * iy2;
         ra00 -= ix6 * iy2;
         ia00 += ix6 * ry2;
         ra10 -= ix7 * iy2;
         ia10 += ix7 * ry2;
         ra00 += rw6 * rz2;
         ia00 += rw6 * iz2;
         ra10 += rw7 * rz2;
         ia10 += rw7 * iz2;
         ra00 -= iw6 * iz2;
         A2[i+12] = ra00;
         ia00 += iw6 * rz2;
         A2[i+12+1] = ia00;
         ra10 -= iw7 * iz2;
         A2[i+12+2] = ra10;
         ia10 += iw7 * rz2;
         A2[i+12+3] = ia10;

         ra00 = A3[i+0];
         ia00 = A3[i+0+1];
         ra10 = A3[i+0+2];
         ia10 = A3[i+0+3];
         ra00 += rx0 * ry3;
         ia00 += rx0 * iy3;
         ra10 += rx1 * ry3;
         ia10 += rx1 * iy3;
         prefA(A3);
         ra00 -= ix0 * iy3;
         ia00 += ix0 * ry3;
         ra10 -= ix1 * iy3;
         ia10 += ix1 * ry3;
         ra00 += rw0 * rz3;
         ia00 += rw0 * iz3;
         ra10 += rw1 * rz3;
         ia10 += rw1 * iz3;
         ra00 -= iw0 * iz3;
         A3[i+0] = ra00;
         ia00 += iw0 * rz3;
         A3[i+0+1] = ia00;
         ra10 -= iw1 * iz3;
         A3[i+0+2] = ra10;
         ia10 += iw1 * rz3;
         A3[i+0+3] = ia10;
         ra00 = A3[i+4];
         ia00 = A3[i+4+1];
         ra10 = A3[i+4+2];
         ia10 = A3[i+4+3];
         ra00 += rx2 * ry3;
         ia00 += rx2 * iy3;
         ra10 += rx3 * ry3;
         ia10 += rx3 * iy3;
         ra00 -= ix2 * iy3;
         ia00 += ix2 * ry3;
         ra10 -= ix3 * iy3;
         ia10 += ix3 * ry3;
         ra00 += rw2 * rz3;
         ia00 += rw2 * iz3;
         ra10 += rw3 * rz3;
         ia10 += rw3 * iz3;
         ra00 -= iw2 * iz3;
         A3[i+4] = ra00;
         ia00 += iw2 * rz3;
         A3[i+4+1] = ia00;
         ra10 -= iw3 * iz3;
         A3[i+4+2] = ra10;
         ia10 += iw3 * rz3;
         A3[i+4+3] = ia10;
         ra00 = A3[i+8];
         ia00 = A3[i+8+1];
         ra10 = A3[i+8+2];
         ia10 = A3[i+8+3];
         ra00 += rx4 * ry3;
         ia00 += rx4 * iy3;
         ra10 += rx5 * ry3;
         ia10 += rx5 * iy3;
         ra00 -= ix4 * iy3;
         ia00 += ix4 * ry3;
         ra10 -= ix5 * iy3;
         ia10 += ix5 * ry3;
         ra00 += rw4 * rz3;
         ia00 += rw4 * iz3;
         ra10 += rw5 * rz3;
         ia10 += rw5 * iz3;
         ra00 -= iw4 * iz3;
         A3[i+8] = ra00;
         ia00 += iw4 * rz3;
         A3[i+8+1] = ia00;
         ra10 -= iw5 * iz3;
         A3[i+8+2] = ra10;
         ia10 += iw5 * rz3;
         A3[i+8+3] = ia10;
         ra00 = A3[i+12];
         ia00 = A3[i+12+1];
         ra10 = A3[i+12+2];
         ia10 = A3[i+12+3];
         ra00 += rx6 * ry3;
         ia00 += rx6 * iy3;
         ra10 += rx7 * ry3;
         ia10 += rx7 * iy3;
         ra00 -= ix6 * iy3;
         ia00 += ix6 * ry3;
         ra10 -= ix7 * iy3;
         ia10 += ix7 * ry3;
         ra00 += rw6 * rz3;
         ia00 += rw6 * iz3;
         ra10 += rw7 * rz3;
         ia10 += rw7 * iz3;
         ra00 -= iw6 * iz3;
         A3[i+12] = ra00;
         ia00 += iw6 * rz3;
         A3[i+12+1] = ia00;
         ra10 -= iw7 * iz3;
         A3[i+12+2] = ra10;
         ia10 += iw7 * rz3;
         A3[i+12+3] = ia10;

         ra00 = A4[i+0];
         ia00 = A4[i+0+1];
         ra10 = A4[i+0+2];
         ia10 = A4[i+0+3];
         ra00 += rx0 * ry4;
         ia00 += rx0 * iy4;
         ra10 += rx1 * ry4;
         ia10 += rx1 * iy4;
         prefA(A4);
         ra00 -= ix0 * iy4;
         ia00 += ix0 * ry4;
         ra10 -= ix1 * iy4;
         ia10 += ix1 * ry4;
         ra00 += rw0 * rz4;
         ia00 += rw0 * iz4;
         ra10 += rw1 * rz4;
         ia10 += rw1 * iz4;
         ra00 -= iw0 * iz4;
         A4[i+0] = ra00;
         ia00 += iw0 * rz4;
         A4[i+0+1] = ia00;
         ra10 -= iw1 * iz4;
         A4[i+0+2] = ra10;
         ia10 += iw1 * rz4;
         A4[i+0+3] = ia10;
         ra00 = A4[i+4];
         ia00 = A4[i+4+1];
         ra10 = A4[i+4+2];
         ia10 = A4[i+4+3];
         ra00 += rx2 * ry4;
         ia00 += rx2 * iy4;
         ra10 += rx3 * ry4;
         ia10 += rx3 * iy4;
         ra00 -= ix2 * iy4;
         ia00 += ix2 * ry4;
         ra10 -= ix3 * iy4;
         ia10 += ix3 * ry4;
         ra00 += rw2 * rz4;
         ia00 += rw2 * iz4;
         ra10 += rw3 * rz4;
         ia10 += rw3 * iz4;
         ra00 -= iw2 * iz4;
         A4[i+4] = ra00;
         ia00 += iw2 * rz4;
         A4[i+4+1] = ia00;
         ra10 -= iw3 * iz4;
         A4[i+4+2] = ra10;
         ia10 += iw3 * rz4;
         A4[i+4+3] = ia10;
         ra00 = A4[i+8];
         ia00 = A4[i+8+1];
         ra10 = A4[i+8+2];
         ia10 = A4[i+8+3];
         ra00 += rx4 * ry4;
         ia00 += rx4 * iy4;
         ra10 += rx5 * ry4;
         ia10 += rx5 * iy4;
         ra00 -= ix4 * iy4;
         ia00 += ix4 * ry4;
         ra10 -= ix5 * iy4;
         ia10 += ix5 * ry4;
         ra00 += rw4 * rz4;
         ia00 += rw4 * iz4;
         ra10 += rw5 * rz4;
         ia10 += rw5 * iz4;
         ra00 -= iw4 * iz4;
         A4[i+8] = ra00;
         ia00 += iw4 * rz4;
         A4[i+8+1] = ia00;
         ra10 -= iw5 * iz4;
         A4[i+8+2] = ra10;
         ia10 += iw5 * rz4;
         A4[i+8+3] = ia10;
         ra00 = A4[i+12];
         ia00 = A4[i+12+1];
         ra10 = A4[i+12+2];
         ia10 = A4[i+12+3];
         ra00 += rx6 * ry4;
         ia00 += rx6 * iy4;
         ra10 += rx7 * ry4;
         ia10 += rx7 * iy4;
         ra00 -= ix6 * iy4;
         ia00 += ix6 * ry4;
         ra10 -= ix7 * iy4;
         ia10 += ix7 * ry4;
         ra00 += rw6 * rz4;
         ia00 += rw6 * iz4;
         ra10 += rw7 * rz4;
         ia10 += rw7 * iz4;
         ra00 -= iw6 * iz4;
         A4[i+12] = ra00;
         ia00 += iw6 * rz4;
         A4[i+12+1] = ia00;
         ra10 -= iw7 * iz4;
         A4[i+12+2] = ra10;
         ia10 += iw7 * rz4;
         A4[i+12+3] = ia10;

         ra00 = A5[i+0];
         ia00 = A5[i+0+1];
         ra10 = A5[i+0+2];
         ia10 = A5[i+0+3];
         ra00 += rx0 * ry5;
         ia00 += rx0 * iy5;
         ra10 += rx1 * ry5;
         ia10 += rx1 * iy5;
         prefA(A5);
         ra00 -= ix0 * iy5;
         ia00 += ix0 * ry5;
         ra10 -= ix1 * iy5;
         ia10 += ix1 * ry5;
         ra00 += rw0 * rz5;
         ia00 += rw0 * iz5;
         ra10 += rw1 * rz5;
         ia10 += rw1 * iz5;
         ra00 -= iw0 * iz5;
         A5[i+0] = ra00;
         ia00 += iw0 * rz5;
         A5[i+0+1] = ia00;
         ra10 -= iw1 * iz5;
         A5[i+0+2] = ra10;
         ia10 += iw1 * rz5;
         A5[i+0+3] = ia10;
         ra00 = A5[i+4];
         ia00 = A5[i+4+1];
         ra10 = A5[i+4+2];
         ia10 = A5[i+4+3];
         ra00 += rx2 * ry5;
         ia00 += rx2 * iy5;
         ra10 += rx3 * ry5;
         ia10 += rx3 * iy5;
         ra00 -= ix2 * iy5;
         ia00 += ix2 * ry5;
         ra10 -= ix3 * iy5;
         ia10 += ix3 * ry5;
         ra00 += rw2 * rz5;
         ia00 += rw2 * iz5;
         ra10 += rw3 * rz5;
         ia10 += rw3 * iz5;
         ra00 -= iw2 * iz5;
         A5[i+4] = ra00;
         ia00 += iw2 * rz5;
         A5[i+4+1] = ia00;
         ra10 -= iw3 * iz5;
         A5[i+4+2] = ra10;
         ia10 += iw3 * rz5;
         A5[i+4+3] = ia10;
         ra00 = A5[i+8];
         ia00 = A5[i+8+1];
         ra10 = A5[i+8+2];
         ia10 = A5[i+8+3];
         ra00 += rx4 * ry5;
         ia00 += rx4 * iy5;
         ra10 += rx5 * ry5;
         ia10 += rx5 * iy5;
         ra00 -= ix4 * iy5;
         ia00 += ix4 * ry5;
         ra10 -= ix5 * iy5;
         ia10 += ix5 * ry5;
         ra00 += rw4 * rz5;
         ia00 += rw4 * iz5;
         ra10 += rw5 * rz5;
         ia10 += rw5 * iz5;
         ra00 -= iw4 * iz5;
         A5[i+8] = ra00;
         ia00 += iw4 * rz5;
         A5[i+8+1] = ia00;
         ra10 -= iw5 * iz5;
         A5[i+8+2] = ra10;
         ia10 += iw5 * rz5;
         A5[i+8+3] = ia10;
         ra00 = A5[i+12];
         ia00 = A5[i+12+1];
         ra10 = A5[i+12+2];
         ia10 = A5[i+12+3];
         ra00 += rx6 * ry5;
         ia00 += rx6 * iy5;
         ra10 += rx7 * ry5;
         ia10 += rx7 * iy5;
         ra00 -= ix6 * iy5;
         ia00 += ix6 * ry5;
         ra10 -= ix7 * iy5;
         ia10 += ix7 * ry5;
         ra00 += rw6 * rz5;
         ia00 += rw6 * iz5;
         ra10 += rw7 * rz5;
         ia10 += rw7 * iz5;
         ra00 -= iw6 * iz5;
         A5[i+12] = ra00;
         ia00 += iw6 * rz5;
         A5[i+12+1] = ia00;
         ra10 -= iw7 * iz5;
         A5[i+12+2] = ra10;
         ia10 += iw7 * rz5;
         A5[i+12+3] = ia10;

         ra00 = A6[i+0];
         ia00 = A6[i+0+1];
         ra10 = A6[i+0+2];
         ia10 = A6[i+0+3];
         ra00 += rx0 * ry6;
         ia00 += rx0 * iy6;
         ra10 += rx1 * ry6;
         ia10 += rx1 * iy6;
         prefA(A6);
         ra00 -= ix0 * iy6;
         ia00 += ix0 * ry6;
         ra10 -= ix1 * iy6;
         ia10 += ix1 * ry6;
         ra00 += rw0 * rz6;
         ia00 += rw0 * iz6;
         ra10 += rw1 * rz6;
         ia10 += rw1 * iz6;
         ra00 -= iw0 * iz6;
         A6[i+0] = ra00;
         ia00 += iw0 * rz6;
         A6[i+0+1] = ia00;
         ra10 -= iw1 * iz6;
         A6[i+0+2] = ra10;
         ia10 += iw1 * rz6;
         A6[i+0+3] = ia10;
         ra00 = A6[i+4];
         ia00 = A6[i+4+1];
         ra10 = A6[i+4+2];
         ia10 = A6[i+4+3];
         ra00 += rx2 * ry6;
         ia00 += rx2 * iy6;
         ra10 += rx3 * ry6;
         ia10 += rx3 * iy6;
         ra00 -= ix2 * iy6;
         ia00 += ix2 * ry6;
         ra10 -= ix3 * iy6;
         ia10 += ix3 * ry6;
         ra00 += rw2 * rz6;
         ia00 += rw2 * iz6;
         ra10 += rw3 * rz6;
         ia10 += rw3 * iz6;
         ra00 -= iw2 * iz6;
         A6[i+4] = ra00;
         ia00 += iw2 * rz6;
         A6[i+4+1] = ia00;
         ra10 -= iw3 * iz6;
         A6[i+4+2] = ra10;
         ia10 += iw3 * rz6;
         A6[i+4+3] = ia10;
         ra00 = A6[i+8];
         ia00 = A6[i+8+1];
         ra10 = A6[i+8+2];
         ia10 = A6[i+8+3];
         ra00 += rx4 * ry6;
         ia00 += rx4 * iy6;
         ra10 += rx5 * ry6;
         ia10 += rx5 * iy6;
         ra00 -= ix4 * iy6;
         ia00 += ix4 * ry6;
         ra10 -= ix5 * iy6;
         ia10 += ix5 * ry6;
         ra00 += rw4 * rz6;
         ia00 += rw4 * iz6;
         ra10 += rw5 * rz6;
         ia10 += rw5 * iz6;
         ra00 -= iw4 * iz6;
         A6[i+8] = ra00;
         ia00 += iw4 * rz6;
         A6[i+8+1] = ia00;
         ra10 -= iw5 * iz6;
         A6[i+8+2] = ra10;
         ia10 += iw5 * rz6;
         A6[i+8+3] = ia10;
         ra00 = A6[i+12];
         ia00 = A6[i+12+1];
         ra10 = A6[i+12+2];
         ia10 = A6[i+12+3];
         ra00 += rx6 * ry6;
         ia00 += rx6 * iy6;
         ra10 += rx7 * ry6;
         ia10 += rx7 * iy6;
         ra00 -= ix6 * iy6;
         ia00 += ix6 * ry6;
         ra10 -= ix7 * iy6;
         ia10 += ix7 * ry6;
         ra00 += rw6 * rz6;
         ia00 += rw6 * iz6;
         ra10 += rw7 * rz6;
         ia10 += rw7 * iz6;
         ra00 -= iw6 * iz6;
         A6[i+12] = ra00;
         ia00 += iw6 * rz6;
         A6[i+12+1] = ia00;
         ra10 -= iw7 * iz6;
         A6[i+12+2] = ra10;
         ia10 += iw7 * rz6;
         A6[i+12+3] = ia10;

         ra00 = A7[i+0];
         ia00 = A7[i+0+1];
         ra10 = A7[i+0+2];
         ia10 = A7[i+0+3];
         ra00 += rx0 * ry7;
         ia00 += rx0 * iy7;
         ra10 += rx1 * ry7;
         ia10 += rx1 * iy7;
         prefA(A7);
         ra00 -= ix0 * iy7;
         ia00 += ix0 * ry7;
         ra10 -= ix1 * iy7;
         ia10 += ix1 * ry7;
         ra00 += rw0 * rz7;
         ia00 += rw0 * iz7;
         ra10 += rw1 * rz7;
         ia10 += rw1 * iz7;
         ra00 -= iw0 * iz7;
         A7[i+0] = ra00;
         ia00 += iw0 * rz7;
         A7[i+0+1] = ia00;
         ra10 -= iw1 * iz7;
         A7[i+0+2] = ra10;
         ia10 += iw1 * rz7;
         A7[i+0+3] = ia10;
         ra00 = A7[i+4];
         ia00 = A7[i+4+1];
         ra10 = A7[i+4+2];
         ia10 = A7[i+4+3];
         ra00 += rx2 * ry7;
         ia00 += rx2 * iy7;
         ra10 += rx3 * ry7;
         ia10 += rx3 * iy7;
         ra00 -= ix2 * iy7;
         ia00 += ix2 * ry7;
         ra10 -= ix3 * iy7;
         ia10 += ix3 * ry7;
         ra00 += rw2 * rz7;
         ia00 += rw2 * iz7;
         ra10 += rw3 * rz7;
         ia10 += rw3 * iz7;
         ra00 -= iw2 * iz7;
         A7[i+4] = ra00;
         ia00 += iw2 * rz7;
         A7[i+4+1] = ia00;
         ra10 -= iw3 * iz7;
         A7[i+4+2] = ra10;
         ia10 += iw3 * rz7;
         A7[i+4+3] = ia10;
         ra00 = A7[i+8];
         ia00 = A7[i+8+1];
         ra10 = A7[i+8+2];
         ia10 = A7[i+8+3];
         ra00 += rx4 * ry7;
         ia00 += rx4 * iy7;
         ra10 += rx5 * ry7;
         ia10 += rx5 * iy7;
         ra00 -= ix4 * iy7;
         ia00 += ix4 * ry7;
         ra10 -= ix5 * iy7;
         ia10 += ix5 * ry7;
         ra00 += rw4 * rz7;
         ia00 += rw4 * iz7;
         ra10 += rw5 * rz7;
         ia10 += rw5 * iz7;
         ra00 -= iw4 * iz7;
         A7[i+8] = ra00;
         ia00 += iw4 * rz7;
         A7[i+8+1] = ia00;
         ra10 -= iw5 * iz7;
         A7[i+8+2] = ra10;
         ia10 += iw5 * rz7;
         A7[i+8+3] = ia10;
         ra00 = A7[i+12];
         ia00 = A7[i+12+1];
         ra10 = A7[i+12+2];
         ia10 = A7[i+12+3];
         ra00 += rx6 * ry7;
         ia00 += rx6 * iy7;
         ra10 += rx7 * ry7;
         ia10 += rx7 * iy7;
         ra00 -= ix6 * iy7;
         ia00 += ix6 * ry7;
         ra10 -= ix7 * iy7;
         ia10 += ix7 * ry7;
         ra00 += rw6 * rz7;
         ia00 += rw6 * iz7;
         ra10 += rw7 * rz7;
         ia10 += rw7 * iz7;
         ra00 -= iw6 * iz7;
         A7[i+12] = ra00;
         ia00 += iw6 * rz7;
         A7[i+12+1] = ia00;
         ra10 -= iw7 * iz7;
         A7[i+12+2] = ra10;
         ia10 += iw7 * rz7;
         A7[i+12+3] = ia10;

      }
      for (i=M8; i < M; i += 2)
      {
         const float rx0 = X[i], ix0 = X[i+1], rw0 = W[i], iw0 = W[i+1];
         A0[i] += rx0 * ry0 - ix0 * iy0 + rw0 * rz0 - iw0 * iz0 ;
         A0[i+1] += ix0 * ry0 + rx0 * iy0 + iw0 * rz0 + rw0 * iz0;
         A1[i] += rx0 * ry1 - ix0 * iy1 + rw0 * rz1 - iw0 * iz1 ;
         A1[i+1] += ix0 * ry1 + rx0 * iy1 + iw0 * rz1 + rw0 * iz1;
         A2[i] += rx0 * ry2 - ix0 * iy2 + rw0 * rz2 - iw0 * iz2 ;
         A2[i+1] += ix0 * ry2 + rx0 * iy2 + iw0 * rz2 + rw0 * iz2;
         A3[i] += rx0 * ry3 - ix0 * iy3 + rw0 * rz3 - iw0 * iz3 ;
         A3[i+1] += ix0 * ry3 + rx0 * iy3 + iw0 * rz3 + rw0 * iz3;
         A4[i] += rx0 * ry4 - ix0 * iy4 + rw0 * rz4 - iw0 * iz4 ;
         A4[i+1] += ix0 * ry4 + rx0 * iy4 + iw0 * rz4 + rw0 * iz4;
         A5[i] += rx0 * ry5 - ix0 * iy5 + rw0 * rz5 - iw0 * iz5 ;
         A5[i+1] += ix0 * ry5 + rx0 * iy5 + iw0 * rz5 + rw0 * iz5;
         A6[i] += rx0 * ry6 - ix0 * iy6 + rw0 * rz6 - iw0 * iz6 ;
         A6[i+1] += ix0 * ry6 + rx0 * iy6 + iw0 * rz6 + rw0 * iz6;
         A7[i] += rx0 * ry7 - ix0 * iy7 + rw0 * rz7 - iw0 * iz7 ;
         A7[i+1] += ix0 * ry7 + rx0 * iy7 + iw0 * rz7 + rw0 * iz7;
      }
      Z += 16;
   }
/*
 * Do remaining columns with NU=1 cleanup
 */
   for (j=N-N8; j; j--, A += lda, Y += 2)
   {
      float *A0=A;
      const register float ry0=Y[0], rz0=Z[0], iy0=Y[1], iz0=Z[1];
      ATL_INT i;
      prefY(Y+16+16-1);
      for (i=0; i < M8; i += 16)
      {
         register float ra00, ia00, ra10, ia10;
         const float rx0=X[i+0], rw0=W[i+0], ix0=X[i+1], iw0=W[i+1], rx1=X[i+2],
                     rw1=W[i+2], ix1=X[i+3], iw1=W[i+3], rx2=X[i+4], rw2=W[i+4],
                     ix2=X[i+5], iw2=W[i+5], rx3=X[i+6], rw3=W[i+6], ix3=X[i+7],
                     iw3=W[i+7], rx4=X[i+8], rw4=W[i+8], ix4=X[i+9], iw4=W[i+9],
                     rx5=X[i+10], rw5=W[i+10], ix5=X[i+11], iw5=W[i+11],
                     rx6=X[i+12], rw6=W[i+12], ix6=X[i+13], iw6=W[i+13],
                     rx7=X[i+14], rw7=W[i+14], ix7=X[i+15], iw7=W[i+15];

         prefX(X);
         ra00 = A0[i+0];
         ia00 = A0[i+0+1];
         ra10 = A0[i+0+2];
         ia10 = A0[i+0+3];
         ra00 += rx0 * ry0;
         ia00 += rx0 * iy0;
         ra10 += rx1 * ry0;
         ia10 += rx1 * iy0;
         prefA(A0);
         ra00 -= ix0 * iy0;
         ia00 += ix0 * ry0;
         ra10 -= ix1 * iy0;
         ia10 += ix1 * ry0;
         prefX(W);
         ra00 += rw0 * rz0;
         ia00 += rw0 * iz0;
         ra10 += rw1 * rz0;
         ia10 += rw1 * iz0;
         ra00 -= iw0 * iz0;
         A0[i+0] = ra00;
         ia00 += iw0 * rz0;
         A0[i+0+1] = ia00;
         ra10 -= iw1 * iz0;
         A0[i+0+2] = ra10;
         ia10 += iw1 * rz0;
         A0[i+0+3] = ia10;
         ra00 = A0[i+4];
         ia00 = A0[i+4+1];
         ra10 = A0[i+4+2];
         ia10 = A0[i+4+3];
         ra00 += rx2 * ry0;
         ia00 += rx2 * iy0;
         ra10 += rx3 * ry0;
         ia10 += rx3 * iy0;
         ra00 -= ix2 * iy0;
         ia00 += ix2 * ry0;
         ra10 -= ix3 * iy0;
         ia10 += ix3 * ry0;
         ra00 += rw2 * rz0;
         ia00 += rw2 * iz0;
         ra10 += rw3 * rz0;
         ia10 += rw3 * iz0;
         ra00 -= iw2 * iz0;
         A0[i+4] = ra00;
         ia00 += iw2 * rz0;
         A0[i+4+1] = ia00;
         ra10 -= iw3 * iz0;
         A0[i+4+2] = ra10;
         ia10 += iw3 * rz0;
         A0[i+4+3] = ia10;
         ra00 = A0[i+8];
         ia00 = A0[i+8+1];
         ra10 = A0[i+8+2];
         ia10 = A0[i+8+3];
         ra00 += rx4 * ry0;
         ia00 += rx4 * iy0;
         ra10 += rx5 * ry0;
         ia10 += rx5 * iy0;
         ra00 -= ix4 * iy0;
         ia00 += ix4 * ry0;
         ra10 -= ix5 * iy0;
         ia10 += ix5 * ry0;
         ra00 += rw4 * rz0;
         ia00 += rw4 * iz0;
         ra10 += rw5 * rz0;
         ia10 += rw5 * iz0;
         ra00 -= iw4 * iz0;
         A0[i+8] = ra00;
         ia00 += iw4 * rz0;
         A0[i+8+1] = ia00;
         ra10 -= iw5 * iz0;
         A0[i+8+2] = ra10;
         ia10 += iw5 * rz0;
         A0[i+8+3] = ia10;
         ra00 = A0[i+12];
         ia00 = A0[i+12+1];
         ra10 = A0[i+12+2];
         ia10 = A0[i+12+3];
         ra00 += rx6 * ry0;
         ia00 += rx6 * iy0;
         ra10 += rx7 * ry0;
         ia10 += rx7 * iy0;
         ra00 -= ix6 * iy0;
         ia00 += ix6 * ry0;
         ra10 -= ix7 * iy0;
         ia10 += ix7 * ry0;
         ra00 += rw6 * rz0;
         ia00 += rw6 * iz0;
         ra10 += rw7 * rz0;
         ia10 += rw7 * iz0;
         ra00 -= iw6 * iz0;
         A0[i+12] = ra00;
         ia00 += iw6 * rz0;
         A0[i+12+1] = ia00;
         ra10 -= iw7 * iz0;
         A0[i+12+2] = ra10;
         ia10 += iw7 * rz0;
         A0[i+12+3] = ia10;

      }
      for (i=M8; i < M; i += 2)
      {
         const float rx0 = X[i], ix0 = X[i+1], rw0 = W[i], iw0 = W[i+1];
         A0[i] += rx0 * ry0 - ix0 * iy0 + rw0 * rz0 - iw0 * iz0 ;
         A0[i+1] += ix0 * ry0 + rx0 * iy0 + iw0 * rz0 + rw0 * iz0;
      }
      Z += 2;
   }
}
